
/* stn41io -- driver for sym41 communication lines, 
	using various GPIO lines. */

/* Character device: 
	read and write for data traffic 
	ioctls for device reset, set/read operation mode

	Thus all gpios used will go through this driver.
*/

#include <linux/init.h>
#include <linux/miscdevice.h>

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */
#include <linux/proc_fs.h> /* /proc entries */
#include <linux/seq_file.h> /* proc entries io */ 
#include <linux/delay.h> 	/* msleep */

#include <linux/interrupt.h>
#include <linux/gpio.h>

#include <linux/signal.h>  /* Signal numbers etc. */
#include <linux/sched.h>  /* process-ident things. */

#include <linux/slab.h>   /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/mutex.h>
#include <linux/kfifo.h>
#include <linux/poll.h>
#include <linux/fcntl.h>  /* O_ACCMODE */
// #include <asm/system.h>   /* cli(), *_flags */
#include <asm/uaccess.h>  /* copy_from/to_user */

#include "stn41_u.h"  /* ioctl commands. */

/* GPIO space */
#include <mach/platform.h>

MODULE_LICENSE("Dual BSD/GPL");

#define A_GBLEN 800

/* GPIO assignments:
	From the 8085 to the R.pi (r.pi read) */ 

#define LABEL_05 "Clock5"
static short gpio_05_Clock5 = 5;   /* out 0 LS165 clock */
#define LABEL_06 "Data5"
static short gpio_06_Data5 = 6;   /* in LS165 data out */
#define LABEL_12 "Le5"
static short gpio_12_Le5= 12;     /* out 1 LS165 latch enable (H-L-H) */
#define LABEL_20 "req5"
static short gpio_20_req5 = 20;   /* out 1 Request clear (L) */
#define LABEL_21 "sta5"
static short gpio_21_sta5 = 21;   /* in Request status */

/* Make an interrupt on gpio21, when this is set at 1 the
	8085 has written a byte and this is ready to be received.
	Put this into a kernel fifo so that we can have it sit there until
	a read request comes along. */ 
static short int irq_21_sta5;

	/* From the R.pi to the 8085 (r.pi write) */
#define LABEL_26 "Ready4"
static short gpio_26_Ready4 = 26; /* in outbound data ready bit */
#define LABEL_13 "Data4"
static short gpio_13_Data4 = 13;  /* out x 4094 data */
#define LABEL_19 "Clock4"
static short gpio_19_Clock4 = 19; /* out 0 4094 clock (L-H) */
#define LABEL_16 "Le4"
static short gpio_16_Le4 = 16;    /* out 0 4094 LE, outbound Data ready set (L-H-L) */

	/* Control and status */
#define LABEL_23 "Reset"
static short gpio_23_Reset = 23; /* out 1 8085 Reset. Usually H, pulse H-L-H to reset the 8085  */
#define LABEL_25 "Mode"
static short gpio_25_Mode = 25;  /* in,  mode in, maps to port 0x0 bit 2 also.  Pulled H with a resistor. Alternate mode is L */

#define THISFILE "stn41io.c"

/* Variables */

/* enforce opening only once. */
static DEFINE_MUTEX(stn41_device_mutex);

#define STN41_FIFOSIZE 1024
static DEFINE_KFIFO(stn41_fifo, char, STN41_FIFOSIZE);
/* fifo becomes a 256-character one. */

/* Declaration of stn41io.c functions */

static int s41_open(struct inode *inode, struct file *filp);
static int s41_close(struct inode *inode, struct file *filp);
static ssize_t s41_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
static ssize_t s41_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
static unsigned int s41_poll(struct file *file, poll_table *wait);
static void s41_exit(void);
static int s41_init(void);

static int s41_opmode_open(struct inode *inode, struct file *file);
static int s41_opmode_show(struct seq_file *f, void *v);

static ssize_t s41_reset_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);

//  add either ioctl or /sys files ...

/* Structure that declares the access functions. */

struct file_operations stn41dev_fops = 
{
	read: s41_read,
	write: s41_write,
	poll: s41_poll,
	open: s41_open,
	release: s41_close
};

/* Misc device structure. */
static struct miscdevice stn41dev_misc_device 
	= {
	.mode = 0666,
	.minor = MISC_DYNAMIC_MINOR,
	.name = "stn41dev",
	.fops = &stn41dev_fops 
  };

/* proc files.  */
static struct file_operations stn41_proc_opmode_fops = 
{
	open: s41_opmode_open,
	read: seq_read,
	llseek: seq_lseek,
	release: single_release
};

static struct file_operations stn41_proc_reset_fops = 
{
	// open: s41_proc_open,
	// release:  s41_proc_close
	write: s41_reset_write
};

	

/* Declaration of the init and exit functions */
module_init(s41_init);
module_exit(s41_exit);


/* use:
	gpio_direction_input(gpio_pb);
	value = gpio_get_value(gpio_pb);

	or: 
	gpio_direction_output(gpio_pb, value);
	gpio_set_value(gpio_pb, value);
*/ 


/* Read the contents of the LS165, no waiting. */
static int pullbyte(void);


/* ISR handler for the gpio21 */
static irqreturn_t irq_21_handler(int irq, void *dev_id, 
	struct pt_regs *regs)
{
	unsigned long flags;
	int pc, item;

	local_irq_save(flags);

	/* gpio21 has gone high, that means there is a byte of data
		to be pulled out of the LS165 */
	item = pullbyte();

	/* Now put this into the fifo */
	pc = kfifo_put(&stn41_fifo, item);

	printk(KERN_INFO "%s(%d): FIFO put %d status %d\n", 
			THISFILE, __LINE__, item, pc);

	local_irq_restore(flags);
	return IRQ_HANDLED;
}

static struct proc_dir_entry *stn41_ppdir;

/* Init: Allocate gpios, create the device entry. */
static int s41_init(void) 
{
	int rx;

	mutex_init(&stn41_device_mutex);

	/* Make the fifo ready to go */
	INIT_KFIFO(stn41_fifo);

	/* Request the gpio lines. */

	rx = gpio_request(gpio_05_Clock5, LABEL_05);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_05_Clock5, LABEL_05, rx); 
		goto nogpio05;
	}

	// static short gpio_05_Clock5 = 5;   /* out 0 LS165 clock */
	rx = gpio_direction_output(gpio_05_Clock5, 0);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_05_Clock5, LABEL_05, rx); 
		goto nogpio06; /* one further */
	}

	rx = gpio_request(gpio_06_Data5, LABEL_06);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_06_Data5, LABEL_06, rx); 
		goto nogpio06;
	}

	// static short gpio_06_Data5 = 6;   /* in LS165 data out */
	rx = gpio_direction_input(gpio_06_Data5);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_06_Data5, LABEL_06, rx); 
		goto nogpio12; /* one further */
	}

	rx = gpio_request(gpio_12_Le5, LABEL_12);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_12_Le5, LABEL_12, rx); 
		goto nogpio12;
	}

	// static short gpio_12_Le5= 12;     /* out 1 LS165 latch enable (H-L-H) */
	rx = gpio_direction_output(gpio_12_Le5, 1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_12_Le5, LABEL_12, rx); 
		goto nogpio20; /* one further */
	}

	rx = gpio_request(gpio_20_req5, LABEL_20);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_20_req5, LABEL_20, rx); 
		goto nogpio20;
	}

	// static short gpio_20_req5 = 20;   /* out 1 Request clear (L) */
	rx = gpio_direction_output(gpio_20_req5, 1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_20_req5, LABEL_20, rx); 
		goto nogpio21; /* one further */
	}

	rx = gpio_request(gpio_21_sta5, LABEL_21);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_21_sta5, LABEL_21, rx); 
		goto nogpio21;
	}

	// static short gpio_21_sta5 = 21;   /* in Request status */
	rx = gpio_direction_input(gpio_21_sta5);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_21_sta5, LABEL_21, rx); 
		goto noirq21; /* one further */
	}

	irq_21_sta5 = gpio_to_irq(gpio_21_sta5); 
	if(irq_21_sta5 < 0)
	{
        printk(KERN_INFO "%s(%d): GPIO %d %s to IRQ mapping failure %d\n",
            THISFILE, __LINE__, gpio_21_sta5, LABEL_21, irq_21_sta5);

		rx = irq_21_sta5;
		goto noirq21;
	}

    printk(KERN_INFO "%s(%d): Mapped int %d from gpio-line %d %s\n",
            THISFILE, __LINE__, irq_21_sta5, gpio_21_sta5, LABEL_21);

    rx = request_irq(irq_21_sta5,
            (irq_handler_t) irq_21_handler,
            IRQF_TRIGGER_RISING, /* IRQF_TRIGGER_FALLING is an alternative */
            LABEL_21,
            "stn41io");

    if(rx < 0)
    {
        printk(KERN_INFO "%s(%d): Irq %d trigger-request on gpio %d %s failure %d\n", 
            THISFILE, __LINE__, irq_21_sta5, gpio_21_sta5, LABEL_21, rx);
		goto noirq21;
    }

	rx = gpio_request(gpio_26_Ready4, LABEL_26);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_26_Ready4, LABEL_26, rx); 
		goto nogpio26;
	}

	// static short gpio_26_Ready4 = 26; /* in outbound data ready bit */
	rx = gpio_direction_input(gpio_26_Ready4);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_26_Ready4, LABEL_26, rx); 
		goto nogpio13; /* one further */
	}

	rx = gpio_request(gpio_13_Data4, LABEL_13);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_13_Data4, LABEL_13, rx); 
		goto nogpio13;
	}

	// static short gpio_13_Data4 = 13;  /* out x 4094 data */
	rx = gpio_direction_output(gpio_13_Data4, 1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_13_Data4, LABEL_13, rx); 
		goto nogpio19; /* one further */
	}

	rx = gpio_request(gpio_19_Clock4, LABEL_19);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_19_Clock4, LABEL_19, rx); 
		goto nogpio19;
	}

	// static short gpio_19_Clock4 = 19; /* out 0 4094 clock (L-H) */
	rx = gpio_direction_output(gpio_19_Clock4, 0);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_19_Clock4, LABEL_19, rx); 
		goto nogpio16; /* one further */
	}

	rx = gpio_request(gpio_16_Le4, LABEL_16);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_16_Le4, LABEL_16, rx); 
		goto nogpio16;
	}

	// static short gpio_16_Le4 = 16;    /* out 0 4094 LE, outbound Data ready set (L-H-L) */
	rx = gpio_direction_output(gpio_16_Le4, 0);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_16_Le4, LABEL_16, rx); 
		goto nogpio23; /* one further */
	}


	/* Control and status */
	rx = gpio_request(gpio_23_Reset, LABEL_23);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_23_Reset, LABEL_23, rx); 
		goto nogpio23;
	}

	// static short gpio_23_Reset = 23; /* out 1 8085 Reset. Usually H, pulse H-L-H to reset the 8085  */
	rx = gpio_direction_output(gpio_23_Reset, 1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_23_Reset, LABEL_23, rx); 
		goto nogpio25; /* one further */
	}

	rx = gpio_request(gpio_25_Mode, LABEL_25);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_25_Mode, LABEL_25, rx); 
		goto nogpio25;
	}

	// static short gpio_25_Mode = 25;  /* in,  mode in, maps to port 0x0 bit 2 also.  Pulled H with a resistor. Alternate mode is L */
	rx = gpio_direction_input(gpio_25_Mode);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_25_Mode, LABEL_25, rx); 
		goto nodevice; /* one further */
	}

	/* Register the device itself. */
	rx = misc_register(&stn41dev_misc_device);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Register misc device failure %d\n", THISFILE, __LINE__, rx); 
		goto nodevice;
	}

	gpio_set_value(gpio_20_req5, 1);

	printk(KERN_INFO "%s(%d): Info: misc device minor number = %d\n", 
		THISFILE, __LINE__, stn41dev_misc_device.minor);

	/* Make two /proc entries, one for mode read and 
		the other for reset the 8085 */ 

	stn41_ppdir = proc_mkdir_mode("stn41", 0555, NULL); 
	proc_create("opmode", 0444, stn41_ppdir, &stn41_proc_opmode_fops); 
	proc_create("reset", 0222, stn41_ppdir, &stn41_proc_reset_fops); 

	return 0;

/* Error exits */
	// remove_proc_entry("opmode", stn41_ppdir); 
	// remove_proc_entry("reset", stn41_ppdir); 
	// remove_proc_entry("stn41", NULL); 
// noproc:
nodevice:
	gpio_free(gpio_25_Mode);
nogpio25:
	gpio_free(gpio_23_Reset);
nogpio23:
	gpio_free(gpio_16_Le4);
nogpio16:
	gpio_free(gpio_19_Clock4);
nogpio19:
	gpio_free(gpio_13_Data4);
nogpio13:
	gpio_free(gpio_26_Ready4);
nogpio26:
	free_irq(irq_21_sta5, "stn41io");
noirq21:
	gpio_free(gpio_21_sta5);
nogpio21:
	gpio_free(gpio_20_req5);
nogpio20:
	gpio_free(gpio_12_Le5);
nogpio12:
	gpio_free(gpio_06_Data5);
nogpio06:
	gpio_free(gpio_05_Clock5);
nogpio05:
	// N-A remove the fifo
// nofifo:
	return(rx);
}



static void s41_exit(void) 
{
	remove_proc_entry("opmode", stn41_ppdir); 
	remove_proc_entry("reset", stn41_ppdir); 
	remove_proc_entry("stn41", NULL); 

	// N-A remove the fifo
	free_irq(irq_21_sta5, "stn41io");

	gpio_free(gpio_05_Clock5);
	gpio_free(gpio_06_Data5);
	gpio_free(gpio_12_Le5);
	gpio_free(gpio_20_req5);
	gpio_free(gpio_21_sta5);
	gpio_free(gpio_26_Ready4);
	gpio_free(gpio_13_Data4);
	gpio_free(gpio_19_Clock4);
	gpio_free(gpio_16_Le4);
	gpio_free(gpio_23_Reset);
	gpio_free(gpio_25_Mode);

	misc_deregister(&stn41dev_misc_device);
	printk(KERN_INFO "%s(%d): Removing stn41dev\n", THISFILE, __LINE__);
}


static int s41_open(struct inode *inode, struct file *filp) 
{
	if(!mutex_trylock(&stn41_device_mutex))
	{
		printk(KERN_INFO "%s(%d): Another process is using the device.\n", 
			THISFILE, __LINE__);
			return -EBUSY;
	}

	try_module_get(THIS_MODULE); 

	return 0;
}

static int s41_close(struct inode *inode, struct file *filp) 
{
	mutex_unlock(&stn41_device_mutex);
	module_put(THIS_MODULE); 
	return 0;
}

/* Transmit a byte. */
static int txabyte(int byte)
{
	int j, rdb;
	unsigned int mask;

	/* Check that the GPIO26 is 0 */
	do
	{
		rdb = gpio_get_value(gpio_26_Ready4); 
		// printk(KERN_INFO "Read g26 = %d\n", rdb); 
	}
	while(rdb > 0); 

	/* For 8 bits:
		bit to GPIO13
		pulse GPIO19 L-H-L
		Send out MSB first 
	*/

	mask = 0x80;
	for(j = 0; j<8; j++)
	{
		gpio_set_value(13, byte & mask ? 1 : 0);
		gpio_set_value(19, 1);
		gpio_set_value(19, 0);
		mask >>= 1;
	}

	/* Pulse GPIO16 L-H-L */
	gpio_set_value(16, 1);
	gpio_set_value(16, 0);

	return(1);
}

/* Make an interrupt on gpio21, when this is set at 1 the
	8085 has written a byte and this is ready to be received.
	Put this into a kernel fifo so that we can have it sit there until
	a read request comes along. */ 

/* When the gpio 21 is high and there is something there, 
	read the data from the LS165 and return that */

static int pullbyte(void)
{
	int j; 
	int bfound, rx;
	unsigned int mask, bacc;

	/* Yes, we have stuff. Hit on the LS165. */

	/* Pulse GPIO12, Latch enable, H-L-H */
	gpio_set_value(12, 1);
	gpio_set_value(12, 0);
	gpio_set_value(12, 1);

	/* Data line has the MSB now. */ 
	bacc = 0; 
	mask = 0x80;
	for(j = 0; j<8; j++)
	{
		rx = gpio_get_value(6); 
		if(rx == 1)
		{
			bacc |= mask;
		}
		/* Shift and pass the rest of them */ 
		mask >>= 1;

		/* Clock */
		gpio_set_value(5, 0);
		gpio_set_value(5, 1);
	}
	bfound = bacc;

	/* Pulse GPIO20 H-L-H to reset. */
	gpio_set_value(20, 1);
	gpio_set_value(20, 0);
	gpio_set_value(20, 1);
	return(bfound); 
}


static ssize_t s41_read(struct file *filp, 
	char *buff_out, size_t count, loff_t *f_pos) 
{ 
	size_t copycount; 
	int kfs;

	/* Get the data from the kernel-fifo */
	if(kfifo_is_empty(&stn41_fifo))
	{
		/* No more to be found. */ 
		return(0);
	}

	kfs = kfifo_to_user(&stn41_fifo, buff_out, count, &copycount);
	/* kfs is zero, or could be -EFAULT if there is something wrong. */

	/* copycount is number of characters copied */
	if(kfs == 0)
	{
		*f_pos += copycount;
		return(copycount);
	}
	else /* Error, return that */
	{
		return(kfs);
	}
}

/* select:  */
static unsigned int s41_poll(struct file *file, poll_table *wait)
{
	unsigned int retnmask = 0;	

	/* If we can read: that is, we have some data in the fifo  */
	if(! kfifo_is_empty(&stn41_fifo))
	{
		retnmask |= POLLIN | POLLRDNORM ;
	}
	
	/* If we can write */
	retnmask |= POLLOUT | POLLWRNORM;

	/* Others: */
	// POLLPRI -- Priority actions
	// POLLHUP -- at end of file received on write (hang-up)
	// POLLERR -- Error occurred
	
	return(retnmask);
}

	
#if(0)
	size_t wcount, j;
	char wbuff[A_GBLEN];
	int hab;

	wcount = 0; 
	for(j = 0; j<count; j++)
	{
		hab = checkbyte();
		if(!hab) break;

		// Block, non-block options:
		wbuff[j] = rxabyte();
		wcount ++;
		if(wcount >= A_GBLEN) break;
	}

	/* Transfering data to user space */ 
	copy_to_user(buff_out, wbuff, wcount);

	/* Changing reading position as best suits */ 
	if(*f_pos == 0) 
	{ 
		*f_pos += wcount; 
		return wcount; 
	} 
	else 
	{ 
		return 0; 
	}
#endif


/* Write -- we come in with some text from the outside to be written.  */

static ssize_t s41_write(struct file *filp, const char *buff_in, size_t count, loff_t *f_pos) 
{
	char rdbuff[A_GBLEN];
	ssize_t j;
	ssize_t rcount;
	int cs;

	rcount = A_GBLEN-1 < count ? A_GBLEN-1 : count;

	cs = copy_from_user(rdbuff, buff_in, rcount);
	if(cs == 0)
	{
		for(j = 0; j<rcount; j++)
		{
			txabyte(rdbuff[j]); 
		}
	}
	else
	{
		rcount = cs;
	}

	rcu_read_unlock(); /* Must use or code will loop ? */ 

	return(rcount); 
}


static ssize_t s41_reset_write(struct file *filp, 
	const char *buf, size_t count, loff_t *f_pos)
{
	size_t locount, rcount;
	char msg[128];
	int cs;

	locount = sizeof(msg);
	memset(msg, 0, locount); 
	cs = copy_from_user(msg, buf, locount); 

	printk(KERN_INFO "%s(%d): Received %d\n", 
			THISFILE, __LINE__, cs);

	// 8085 Reset. Usually H, pulse H-L-H to reset the 8085 

	gpio_set_value(gpio_23_Reset, 1);
	gpio_set_value(gpio_23_Reset, 0);
	msleep(50);    /* give the 8085 50 ms to react. */ 
	gpio_set_value(gpio_23_Reset, 1);

	/* printk(KERN_INFO "%s(%d): Received %s %d\n", 
			THISFILE, __LINE__, buf, cs); */ 
	rcount = strlen(buf);

	return(rcount);
}

/* read the value of gpio 25 and return the line level seen. */
static int s41_opmode_show(struct seq_file *f, void *v)
{
	int rdb;

	rdb = gpio_get_value(gpio_25_Mode); 
	seq_printf(f, "%d\n", rdb); 
	return(0);
}

static int s41_opmode_open(struct inode *inode, struct file *file)
{
	return single_open(file, s41_opmode_show, (void *) 0x1234); 
}



