
/* stn41io -- driver for sym41 communication lines, 
	using various GPIO lines. */

/* Character device: 
	read and write for data traffic 
	ioctls for device reset, set/read operation mode

	Thus all gpios used will go through this driver.
*/

#include <linux/init.h>
#include <linux/miscdevice.h>

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */

#include <linux/interrupt.h>
#include <linux/gpio.h>

#include <linux/signal.h>  /* Signal numbers etc. */
#include <linux/sched.h>  /* process-ident things. */

#include <linux/slab.h>   /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>  /* O_ACCMODE */
// #include <asm/system.h>   /* cli(), *_flags */
#include <asm/uaccess.h>  /* copy_from/to_user */

#include "stn41_u.h"  /* ioctl commands. */

/* GPIO space */
#include <mach/platform.h>

MODULE_LICENSE("Dual BSD/GPL");

#define A_GBLEN 800

/* GPIO assignments:
	From the 8085 to the R.pi (r.pi read) */ 

#define LABEL_05 "Clock5"
static short gpio_05_Clock5 = 5;   /* out 0 LS165 clock */
#define LABEL_06 "Data5"
static short gpio_06_Data5 = 6;   /* in LS165 data out */
#define LABEL_12 "Le5"
static short gpio_12_Le5= 12;     /* out 1 LS165 latch enable (H-L-H) */
#define LABEL_20 "req5"
static short gpio_20_req5 = 20;   /* out 1 Request clear (L) */
#define LABEL_21 "sta5"
static short gpio_21_sta5 = 21;   /* in Request status */

	/* From the R.pi to the 8085 (r.pi write) */
#define LABEL_26 "Ready4"
static short gpio_26_Ready4 = 26; /* in outbound data ready bit */
#define LABEL_13 "Data4"
static short gpio_13_Data4 = 13;  /* out x 4094 data */
#define LABEL_19 "Clock4"
static short gpio_19_Clock4 = 19; /* out 0 4094 clock (L-H) */
#define LABEL_16 "Le4"
static short gpio_16_Le4 = 16;    /* out 0 4094 LE, outbound Data ready set (L-H-L) */

	/* Control and status */
#define LABEL_23 "Reset"
static short gpio_23_Reset = 23; /* out 1 8085 Reset. Usually H, pulse H-L-H to reset the 8085  */
#define LABEL_25 "Mode"
static short gpio_25_Mode = 25;  /* in,  mode in, maps to port 0x0 bit 2 also.  Pulled H with a resistor. Alternate mode is L */

#define THISFILE "stn41io.c"

/* Declaration of stn41io.c functions */

static int s41_open(struct inode *inode, struct file *filp);
static int s41_release(struct inode *inode, struct file *filp);
static ssize_t s41_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
static ssize_t s41_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
static void s41_exit(void);
static int s41_init(void);

//  add either ioctl or /sys files ...

/* Structure that declares the access functions. */

struct file_operations stn41dev_fops = 
{
  read: s41_read,
  write: s41_write,
  open: s41_open,
  release: s41_release
};

/* Misc device structure. */
static struct miscdevice stn41dev_misc_device 
	= {
	.mode = 0666,
	.minor = MISC_DYNAMIC_MINOR,
	.name = "stn41dev",
	.fops = &stn41dev_fops 
  };

/* Declaration of the init and exit functions */
module_init(s41_init);
module_exit(s41_exit);

/* enforce opening only once. */
static int s41_open_count = 0;

/* use:
	gpio_direction_input(gpio_pb);
	value = gpio_get_value(gpio_pb);

	or: 
	gpio_direction_output(gpio_pb, value);
	gpio_set_value(gpio_pb, value);
*/ 


/* Init: Allocate gpios, create the device entry. */
static int s41_init(void) 
{
	int rx;

	/* Request the gpio lines. */

	rx = gpio_request(gpio_05_Clock5, LABEL_05);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_05_Clock5, LABEL_05, rx); 
		goto nogpio05;
	}

	// static short gpio_05_Clock5 = 5;   /* out 0 LS165 clock */
	rx = gpio_direction_output(gpio_05_Clock5, 0);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_05_Clock5, LABEL_05, rx); 
		goto nogpio06; /* one further */
	}

	rx = gpio_request(gpio_06_Data5, LABEL_06);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_06_Data5, LABEL_06, rx); 
		goto nogpio06;
	}

	// static short gpio_06_Data5 = 6;   /* in LS165 data out */
	rx = gpio_direction_input(gpio_06_Data5);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_06_Data5, LABEL_06, rx); 
		goto nogpio12; /* one further */
	}

	rx = gpio_request(gpio_12_Le5, LABEL_12);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_12_Le5, LABEL_12, rx); 
		goto nogpio12;
	}

	// static short gpio_12_Le5= 12;     /* out 1 LS165 latch enable (H-L-H) */
	rx = gpio_direction_output(gpio_12_Le5, 1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_12_Le5, LABEL_12, rx); 
		goto nogpio20; /* one further */
	}

	rx = gpio_request(gpio_20_req5, LABEL_20);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_20_req5, LABEL_20, rx); 
		goto nogpio20;
	}

	// static short gpio_20_req5 = 20;   /* out 1 Request clear (L) */
	rx = gpio_direction_output(gpio_20_req5, 1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_20_req5, LABEL_20, rx); 
		goto nogpio21; /* one further */
	}

	rx = gpio_request(gpio_21_sta5, LABEL_21);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_21_sta5, LABEL_21, rx); 
		goto nogpio21;
	}

	// static short gpio_21_sta5 = 21;   /* in Request status */
	rx = gpio_direction_input(gpio_21_sta5);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_21_sta5, LABEL_21, rx); 
		goto nogpio26; /* one further */
	}

	rx = gpio_request(gpio_26_Ready4, LABEL_26);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_26_Ready4, LABEL_26, rx); 
		goto nogpio26;
	}

	// static short gpio_26_Ready4 = 26; /* in outbound data ready bit */
	rx = gpio_direction_input(gpio_26_Ready4);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_26_Ready4, LABEL_26, rx); 
		goto nogpio13; /* one further */
	}

	rx = gpio_request(gpio_13_Data4, LABEL_13);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_13_Data4, LABEL_13, rx); 
		goto nogpio13;
	}

	// static short gpio_13_Data4 = 13;  /* out x 4094 data */
	rx = gpio_direction_output(gpio_13_Data4, 1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_13_Data4, LABEL_13, rx); 
		goto nogpio19; /* one further */
	}

	rx = gpio_request(gpio_19_Clock4, LABEL_19);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_19_Clock4, LABEL_19, rx); 
		goto nogpio19;
	}

	// static short gpio_19_Clock4 = 19; /* out 0 4094 clock (L-H) */
	rx = gpio_direction_output(gpio_19_Clock4, 0);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_19_Clock4, LABEL_19, rx); 
		goto nogpio16; /* one further */
	}

	rx = gpio_request(gpio_16_Le4, LABEL_16);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_16_Le4, LABEL_16, rx); 
		goto nogpio16;
	}

	// static short gpio_16_Le4 = 16;    /* out 0 4094 LE, outbound Data ready set (L-H-L) */
	rx = gpio_direction_output(gpio_16_Le4, 0);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_16_Le4, LABEL_16, rx); 
		goto nogpio23; /* one further */
	}


	/* Control and status */
	rx = gpio_request(gpio_23_Reset, LABEL_23);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_23_Reset, LABEL_23, rx); 
		goto nogpio23;
	}

	// static short gpio_23_Reset = 23; /* out 1 8085 Reset. Usually H, pulse H-L-H to reset the 8085  */
	rx = gpio_direction_output(gpio_23_Reset, 1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_23_Reset, LABEL_23, rx); 
		goto nogpio25; /* one further */
	}

	rx = gpio_request(gpio_25_Mode, LABEL_25);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s request failure %d\n", 
			THISFILE, __LINE__, gpio_25_Mode, LABEL_25, rx); 
		goto nogpio25;
	}

	// static short gpio_25_Mode = 25;  /* in,  mode in, maps to port 0x0 bit 2 also.  Pulled H with a resistor. Alternate mode is L */
	rx = gpio_direction_input(gpio_25_Mode);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d %s output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_25_Mode, LABEL_25, rx); 
		goto nodevice; /* one further */
	}

	/* Register the device itself. */
	rx = misc_register(&stn41dev_misc_device);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Register misc device failure %d\n", THISFILE, __LINE__, rx); 
		goto nodevice;
	}

	printk(KERN_INFO "%s(%d): Info: misc device minor number = %d\n", 
		THISFILE, __LINE__, stn41dev_misc_device.minor);

	return 0;

/* Error exits */
nodevice:
	gpio_free(gpio_25_Mode);
nogpio25:
	gpio_free(gpio_23_Reset);
nogpio23:
	gpio_free(gpio_16_Le4);
nogpio16:
	gpio_free(gpio_19_Clock4);
nogpio19:
	gpio_free(gpio_13_Data4);
nogpio13:
	gpio_free(gpio_26_Ready4);
nogpio26:
	gpio_free(gpio_21_sta5);
nogpio21:
	gpio_free(gpio_20_req5);
nogpio20:
	gpio_free(gpio_12_Le5);
nogpio12:
	gpio_free(gpio_06_Data5);
nogpio06:
	gpio_free(gpio_05_Clock5);
nogpio05:
	return(rx);
}



static void s41_exit(void) 
{
	gpio_free(gpio_05_Clock5);
	gpio_free(gpio_06_Data5);
	gpio_free(gpio_12_Le5);
	gpio_free(gpio_20_req5);
	gpio_free(gpio_21_sta5);
	gpio_free(gpio_26_Ready4);
	gpio_free(gpio_13_Data4);
	gpio_free(gpio_19_Clock4);
	gpio_free(gpio_16_Le4);
	gpio_free(gpio_23_Reset);
	gpio_free(gpio_25_Mode);

	misc_deregister(&stn41dev_misc_device);
	printk(KERN_INFO "%s(%d): Removing stn41dev\n", THISFILE, __LINE__);
}


static int s41_open(struct inode *inode, struct file *filp) 
{
	if(s41_open_count)
		return -EBUSY;

	s41_open_count ++;
	try_module_get(THIS_MODULE); 

	return 0;
}

static int s41_release(struct inode *inode, struct file *filp) 
{
	s41_open_count --;
	module_put(THIS_MODULE);
	return 0;
}

/* Transmit a byte. */
int txabyte(int byte)
{
	int j, rdb;
	unsigned int mask;

	/* Check that the GPIO26 is 0 */
	do
	{
		rdb = gpio_get_value(26); 
		// printk(KERN_INFO "Read g26 = %d\n", rdb); 
	}
	while(rdb > 0); 

	/* For 8 bits:
		bit to GPIO13
		pulse GPIO19 L-H-L
		Send out MSB first 
	*/

	mask = 0x80;
	for(j = 0; j<8; j++)
	{
		gpio_set_value(13, byte & mask ? 1 : 0);
		gpio_set_value(19, 1);
		gpio_set_value(19, 0);
		mask >>= 1;
	}

	/* Pulse GPIO16 L-H-L */
	gpio_set_value(16, 1);
	gpio_set_value(16, 0);

	return(1);
}

/* Receive a byte. */
int checkbyte(void)
{
	/* Check GPIO21, do we have any data? */
	return gpio_get_value(21); 
}

int rxabyte(void)
{
	int j, hasd; 
	int bfound, rx;
	unsigned int mask, bacc;

	/* To be sure... */
	gpio_set_value(20, 1);

	do
	{
		/* Check GPIO21, do we have any data? */
		hasd  = gpio_get_value(21); 
		// printk(KERN_INFO "Read g26 = %d\n", hasd); 
	}
	while(hasd == 0); 

	/* Yes, we have stuff. Hit on the LS165. */

	/* Pulse GPIO12, Latch enable, H-L-H */
	gpio_set_value(12, 1);
	gpio_set_value(12, 0);
	gpio_set_value(12, 1);

	/* Data line has the MSB now. */ 
	bacc = 0; 
	mask = 0x80;
	for(j = 0; j<8; j++)
	{
		rx = gpio_get_value(6); 
		if(rx == 1)
		{
			bacc |= mask;
		}
		/* Shift and pass the rest of them */ 
		mask >>= 1;

		/* Clock */
		gpio_set_value(5, 0);
		gpio_set_value(5, 1);
	}
	bfound = bacc;

	/* Pulse GPIO20 H-L-H to reset. */
	gpio_set_value(20, 1);
	gpio_set_value(20, 0);
	gpio_set_value(20, 1);
	return(bfound); 
}


static ssize_t s41_read(struct file *filp, char *buff_out, size_t count, loff_t *f_pos) 
{ 
	size_t wcount, j;
	char wbuff[A_GBLEN];
	int hab;

	wcount = 0; 
	for(j = 0; j<count; j++)
	{
		hab = checkbyte();
		if(!hab) break;

		// Block, non-block options:
		wbuff[j] = rxabyte();
		wcount ++;
		if(wcount >= A_GBLEN) break;
	}

	/* Transfering data to user space */ 
	copy_to_user(buff_out, wbuff, wcount);

	/* Changing reading position as best suits */ 
	if(*f_pos == 0) 
	{ 
		*f_pos += wcount; 
		return wcount; 
	} 
	else 
	{ 
		return 0; 
	}
}


/* Write -- we come in with some text from the outside to be written.  */

static ssize_t s41_write(struct file *filp, const char *buff_in, size_t count, loff_t *f_pos) 
{
	char rdbuff[A_GBLEN];
	size_t j, rcount;

	rcount = A_GBLEN-1 < count ? A_GBLEN-1 : count;

	copy_from_user(rdbuff, buff_in, rcount);

	for(j = 0; j<rcount; j++)
	{
		txabyte(rdbuff[j]); 
	}

	rcu_read_unlock(); /* Must use or code will loop ? */ 

	return(rcount); 
}



