
/* mgu.c -- start/stop/run the motors. */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <poll.h>
#include <signal.h>
#include <math.h>

#include <utilib.h>
#include "mgu.h"

/* Write a value to one if the gpio files */
int writeto(char *file, char *value)
{
    FILE *fp;
    fp = fopen(file, "w");
    if(!fp) return(-1);

    fprintf(fp, "%s\n", value);
    fclose(fp);
    return(strlen(value));
}

/* Write nvalue to the gpio number value file to set it on or off. */ 
int setgpio(int gpiono, int nvalue)
{
	char thefile[256];
    FILE *fp;
	snprintf(thefile, 256, "/sys/class/gpio/gpio%d/value", gpiono); 
    fp = fopen(thefile, "w");
    if(!fp) return(-1);

    fprintf(fp, "%d\n", nvalue);
    fclose(fp);
    return(nvalue);
}



/* Read current state of one of the gpio files */
int readfrom(char *file, char *readval, size_t rvmax)
{
    FILE *fp;

    fp = fopen(file, "r");
    if(!fp) return(-1);
    fgets(readval, rvmax, fp);

    fclose(fp);
    return(strlen(readval));
}

/* Left motor:						Right motor:
	G17=0 stop 1 go					G10=0 stop 1 go
    G04=0 back 1 forward            G25=0 back 1 forward
*/ 

void mstop()
{
	setgpio(17, 0);
	setgpio(10, 0);
}

void mimpL(int duration, int direction)
{
	setgpio(4, direction); 
	setgpio(17, 1);
	delay(duration); 
	setgpio(17, 0);
}

void mimpR(int duration, int direction)
{
	setgpio(25, direction); 
	setgpio(10, 1);
	delay(duration); 
	setgpio(10, 0);
}

void mimpLR(int durL, int durR, int direction)
{
	int durmin, ddu, stop1, stop2;
	if(durL < durR)
	{
		durmin = durL;
		ddu = durR-durL;	
		stop1 = 17; /* Left stops first */
		stop2 = 10;
	}
	else
	{
		durmin = durR;
		ddu = durL-durR;	
		stop1 = 10;
		stop2 = 17; /* Left stops last */
	}

	setgpio(4, direction); 
	setgpio(25, direction); 
	setgpio(17, 1);
	setgpio(10, 1);

	delay(durmin); 
	setgpio(stop1, 0);
	if(ddu) delay(ddu);
	setgpio(stop2, 0);
}


void mimpLs(int duration)
{
	int dir;
	dir = 1;
	if(duration < 0) dir = 0;

	setgpio(4, dir); 
	setgpio(17, 1);
	delay(abs(duration)); 
	setgpio(17, 0);
}

void mimpRs(int duration)
{
	int dir;
	dir = 1;
	if(duration < 0) dir = 0;
	setgpio(25, dir); 
	setgpio(10, 1);
	delay(abs(duration)); 
	setgpio(10, 0);
}


void mimpLRs(int durL, int durR)
{
	int durmin, ddu, stop1, stop2;
	int aimL, aimR, adurL, adurR;

	adurL = abs(durL);
	adurR = abs(durR);
	aimL = 1; if(durL < 0) aimL = 0; 
	aimR = 1; if(durR < 0) aimR = 0; 

	if(adurL < adurR)
	{
		durmin = adurL;
		ddu = adurR-adurL;	
		stop1 = 17; /* Left stops first */
		stop2 = 10;
	}
	else
	{
		durmin = adurR;
		ddu = adurL-adurR;	
		stop1 = 10;
		stop2 = 17; /* Left stops last */
	}

	setgpio(4, aimL); 
	setgpio(25, aimR); 
	setgpio(17, 1);
	setgpio(10, 1);

	delay(durmin); 
	setgpio(stop1, 0);
	if(ddu) delay(ddu);
	setgpio(stop2, 0);
}


int go_straight(int ireq, int *plt, int *prt)
{

	/* Logic:
		move L and R wheel, making sure they generate the 
		same number of pulses. Never mind which direction 
		we are facing. */ 
	
	struct pollfd fdset[5];
	int leftfd, rightfd; 
	int rc;
	char rdbuf[20];
	int c;
	int odo;
	int lco, rco;
	int nreq;

	if(ireq == 0) return(0); 

	if(ireq < 0)
	{
		nreq = -ireq;
		setgpio(RDIR, GBACK);
		setgpio(LDIR, GBACK);
	}
	else
	{
		nreq = ireq;
		setgpio(RDIR, GFWD);
		setgpio(LDIR, GFWD);
	}

	lco = rco = odo = 0; 

	leftfd = open("/sys/class/gpio/gpio8/value", O_RDONLY | O_NONBLOCK); 
	rightfd = open("/sys/class/gpio/gpio7/value", O_RDONLY | O_NONBLOCK); 

	/* loop */
	for(;;)
	{
		memset( (void *) fdset, 0, 3 * sizeof(struct pollfd));
		fdset[0].fd = leftfd;
		fdset[0].events = POLLPRI;
		fdset[0].revents = 0;
		fdset[1].fd = rightfd;
		fdset[1].events = POLLPRI;
		fdset[1].revents = 0;
		fdset[2].fd = fileno(stdin);
		fdset[2].events = POLLIN;
		fdset[2].revents = 0;

		rc = poll(fdset, 2, 10);

		if(rc < 0 && errno != 4)
		{
			printf("%s(%d): Poll error %d\n", __FILE__, __LINE__, errno);
			break;
		}


		if(rc == 0)
		{
			/* printf("%s(%d): Time-check: Rq=%d Odo=%d L=%d R=%d\n", 
				__FILE__, __LINE__, nreq*2, odo, lco, rco ); */ 

			if(odo > nreq*2) break; 

			if(lco < rco) { setgpio(LGO, 1); }
			else if(lco > rco) { setgpio(RGO, 1); }
			else 
			{
				setgpio(LGO, 1);
				setgpio(RGO, 1);
			}
			continue;
		}

		/* Left wheel moved */
		if(fdset[0].revents & POLLPRI)
		{
			read(fdset[0].fd, rdbuf, 2); 
			lco ++;
			odo ++;
			/* printf("%s(%d): Left Moved now L=%d Odo=%d\n", 
				__FILE__, __LINE__, lco, odo ); */ 
			setgpio(LGO, 0);
		}

		/* right wheel moved */
		if(fdset[1].revents & POLLPRI)
		{
			read(fdset[1].fd, rdbuf, 2); 
			c = rdbuf[0];
			rco ++;
			odo ++;
			/* printf("%s(%d): Right Moved now R=%d Odo=%d\n", 
				__FILE__, __LINE__, rco, odo ); */ 
			setgpio(RGO, 0);
		}

		/* stdin */
		if(fdset[2].revents & POLLIN)
		{
			read(fdset[2].fd, rdbuf, 1); 
			c = rdbuf[0];
			/* printf("%s(%d): stdin hit %d %c\n", __FILE__, __LINE__, c,  c); */ 
			if('q' == c) break;
			if('Q' == c) break;
		}
	}

	setgpio(LGO, 0);
	setgpio(RGO, 0);

	close(leftfd);
	close(rightfd);

	*plt = lco;
	*prt = rco;

	return(odo);
}


